--[[
Copyright (c) 2009 Vincent Pelletier

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Note:
    This AddOn's source code is specifically designed to work with
    World of Warcraft's interpreted AddOn system.
    You have an implicit license to use this AddOn with these facilities
    since that is its designated purpose as per:
    http://www.fsf.org/licensing/licenses/gpl-faq.html#InterpreterIncompat
--]]

SyncFriends = LibStub("AceAddon-3.0"):NewAddon("SyncFriends", "AceConsole-3.0",
  "AceHook-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("SyncFriends", true)

local function copyTable(...)
    --[[
      Iterate over given parameters and returns the result of successive table
      updates.
      Does not modify any of parameters.
      XXX: Does a first-level copy: subtables will not be duplicated.
    --]]
    local result = {}
    for i=1, select("#", ...) do
        for key, value in pairs(select(i, ...)) do
            result[key] = value
        end
    end
    return result
end

local function tchelper(first, rest)
  return first:upper()..rest:lower()
end

local GLOBAL_SCOPE = "Global"
local CHAR_SCOPE = "Char"
local function addGetSet(class, scope_map, property)
    --[[
      Generate accessors for given property, and update option maps given in
      scope_map to use those accessors.
      Those accessors store/retrieve data to/from their own scope in an AceDB
      storage.
      Also creates a synthetic accessor which takes care of property
      precedence.

      @param class Addon object to create accessors on
      @param scope_map List of {scope, option_map} in decreasing precedence
        order
      @param property Name of the property

      Alters: scope_map content maps, class

      TODO: Check parameter sanity (especialy "property")
      XXX: AceDB storage is expected to be accessible with "class.db".
      TODO: cleanup local namespace upon exit, as it will survive (we bind
      functions referencing local namespace, so it won't be garbage-collected)
      and might become a memory hog if used at a big scale.
    --]]

    -- Create individual getters
    local getter_list = {}
    local titlecased_property = string.gsub(property, "_?(%a)([%w]*)",
        tchelper)
    for _, scope in pairs(scope_map) do
        local scope_name = scope[1]
        local scope_option = scope[2]
        local scoped_name = scope_name..titlecased_property
        local getter_name = "get"..scoped_name
        local setter_name = "set"..scoped_name
        local scope_attribute = string.lower(scope_name)

        class[setter_name] = function(info, value)
            class.db[scope_attribute][property] = value
        end
        class[getter_name] = function(info)
            return class.db[scope_attribute][property]
        end

        scope_option.set = class[setter_name]
        scope_option.get = class[getter_name]
        table.insert(getter_list, getter_name)
    end

    -- Create synthetic getter
    class["get"..titlecased_property] = function()
        for _, getter in pairs(getter_list) do
            result = class[getter]()
            if result ~= nil then
                break
            end
        end
        return result
    end
end

local function overridableOptionGroupArgs(option, desc, class, property)
    --[[
      Create 2 copies of given option and link them with (auto-generated)
      property accessors.
      That property will be stored at 2 scopes:
      - Character scope
      - Global scope
      The former overriding setting from the latter.

      @param option Option definition
      @param desc Option description
      @param class Addon object to create option for
      @param property Name of the property

      TODO: Support more than just "toggle" option type.
      XXX: hardcodes the list of scopes
      XXX: hardcodes the way scopes are nested in option_map
    --]]

    local player_option = copyTable(option, {
        name = L["Character override"],
        width = "full",
        order = 2,
    })
    local global_option = copyTable(option, {
        name = L["Global setting"],
        width = "full",
        order = 1,
    })
    local args = {
        desc = {
            name = desc,
            type = "description",
            order = 0,
        },
        player = player_option,
        global = global_option,
    }

    local needs_getset = false
    if option.type == "toggle" then
        player_option["tristate"] = true
        needs_getset = true
    end

    if needs_getset then
        addGetSet(class,
            {{CHAR_SCOPE, player_option}, {GLOBAL_SCOPE, global_option}},
            property)
    end
    return args
end

local function slashCommand(option, class, function_name, argc)
    --[[
      Transforms an option into a slash command with automated parameter
      parsing.

      @param option Option definition to alter
      @param class Addon object defining command method
      @param function_name Name of the function to call on class
      @param argc (opt) The number of arguments to parse

      Alters: option

      TODO: Check parameter sanity better
      XXX: It is dirty to use options for this purpose
    --]]
    local dummy_setter
    option.guiHidden = true
    option.type = "input"
    if argc == 0 then
        -- Don't bother wrapping, we are not expecting parameters.
        dummy_setter = function_name
    else
        if not argc then
            argc = 1e9 -- XXX: This seems to mean "unlimited" for GetArgs.
        end
        dummy_setter = function(info, input)
            class[function_name](class, class:GetArgs(input, argc, 1))
        end
    end
    option.set = dummy_setter
    return option
end

local function dynamicGroup(option, class, method_name)
    --[[
      Returns class[method_name]
      when "args" key is accessed.
      This makes it possible for an option group to be dynamicaly generated.

      @param option Option definition
      @param class Class to call method from
      @param method_name Name of the method to call upon "args" attribute
        access

      Alters: option

      XXX: There should be a better way to do this...
      TODO: Check parameter sanity
    --]]
    local metatable = {
        __index = function(_, key)
            local result
            if key == "args" then
                result = class[method_name](class)
            end
            return result
        end
    }
    setmetatable(option, metatable)
    return option
end

-- This "class" is merely used as a separate namespace to avoid adding those
-- functions to SyncFriends. It is used as a handler for actions defined in
-- ui_sync_pool_action_option
local UISyncPoolOptionHandlers = {}

local options = {
    name = "SyncFriends",
    handler = SyncFriends,
    type = "group",
    childGroups = "tab",
    args = {
        pool = {
            name = L["Sync data"],
            type = "group",
            args = {
                skip = slashCommand({
                        name = L["skip <playerName>"],
                        desc = L["Skip given player name in sync"],
                    },
                    SyncFriends, "skipPlayerName", 1),
                register = slashCommand({
                        name = L["register <playerName> <add|remove>"],
                        desc = L["Register playerName for addition "..
                                 "(default) or removal"],
                    },
                    SyncFriends, "registerFromCommand", 2),
                forget = slashCommand({
                        name = L["forget <playerName>"],
                        desc = L["Forget about playerName (but don't mark "..
                                 "him for removal)"],
                    },
                    SyncFriends, "forget", 1),
                sync = {
                    name = L["Sync"],
                    desc = L["Manually trigger a synchronisation with sync"..
                             " pool"],
                    type = "execute",
                    func = "importExport",
                },
                dump = {
                    name = L["Dump"],
                    desc = L["Dumps sync pool to default chat window"],
                    type = "execute",
                    func = "dump",
                    guiHidden = true,
                },
                flush = {
                    name = L["Flush"],
                    desc = L["Empties the content of the sync pool"],
                    func = "flushPool",
                    type = "execute",
                    confirm = true,
                },
                sync_pool = dynamicGroup({
                        name = L["Friends"],
                        type = "group",
                        handler = UISyncPoolOptionHandlers,
                        cmdHidden = true,
                    },
                    SyncFriends, "getUISyncPoolOption"),
            },
        },
        options = {
            name = L["Options"],
            type = "group",
            args = {
                auto_import = {
                    name = L["Auto-import"],
                    inline = true,
                    type = "group",
                    args = overridableOptionGroupArgs({
                            type = "toggle",
                        },
                        L["Whether friends should be imported upon startup"],
                        SyncFriends, "auto_import"),
                },
                auto_export = {
                    name = L["Auto-export"],
                    inline = true,
                    type = "group",
                    args = overridableOptionGroupArgs({
                            type = "toggle",
                        },
                        L["Whether friends should be made known to "..
                          "synchronisation upon startup, addition and "..
                          "removal"],
                        SyncFriends, "auto_export"),
                },
            },
        },
    }
}

LibStub("AceConfig-3.0"):RegisterOptionsTable("SyncFriends", options,
    {"sf", "syncfriends"})
LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SyncFriends")

-- XXX: Should be fetched from WoW API
local MAX_FRIEND_COUNT = 50

local ADD_ACTION = 1
local REMOVE_ACTION = 2
local SKIP_ACTION = 3

local ACTION_MAP = {
    [ADD_ACTION] = function(playerName, friend_set)
        local result = nil
        if not friend_set[playerName] then
            if #(friend_set) >= MAX_FRIEND_COUNT then
                result = string.format(L["You have reached the maximum "..
                  "number of friends, could not add %s"], playerName)
            else
                AddFriend(playerName)
                friend_set[playerName] = true
                result = string.format(L["Added %s"], playerName)
            end
        end
        return result
    end,
    [REMOVE_ACTION] = function(playerName, friend_set)
        local result = nil
        if friend_set[playerName] then
            RemoveFriend(playerName)
            friend_set[playerName] = nil
            result = string.format(L["Removed %s"], playerName)
        end
        return result
    end,
    [SKIP_ACTION] = function() end,
}

ADD_ACTION_NAME = L["addition"]
REMOVE_ACTION_NAME = L["removal"]
SKIP_ACTION_NAME = L["skipping"]

local ACTION_NAME_MAP = {
    [ADD_ACTION] = "|cff00ff00"..ADD_ACTION_NAME.."|r",
    [REMOVE_ACTION] = "|cffff0000"..REMOVE_ACTION_NAME.."|r",
    [SKIP_ACTION] = "|cff555555"..SKIP_ACTION_NAME.."|r",
}

function SyncFriends:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("SyncFriendsDB")
end

function SyncFriends:OnEnable()
    if self:getAutoImport() == nil and self:getAutoExport() == nil then
        self:Print(L["SyncFriends must be configured before it does anything"])
    else
        self:importExport()
    end
    self:SecureHook("AddFriend")
    self:RawHook("RemoveFriend", true)
end

function SyncFriends:OnDisable()
    self:UnhookAll()
end

function SyncFriends:isFriend(playerName)
    local result = false
    for x = 1, GetNumFriends() do
        if GetFriendInfo(x) == playerName then
            result = true
            break
        end
    end
    return result
end

function SyncFriends:AddFriend(playerName)
    -- After-hook
    -- Register friend for sync iif the addition was successful
    if self.isFriend(playerName) then
        self:storeAction(playerName, ADD_ACTION, false)
    end
end

function SyncFriends:RemoveFriend(playerNameOrID)
    -- Raw hook, because of friend name lookup and actual removal check
    -- register for deletion iif removal was successful
    if type(playerNameOrID) == "number" then
        playerName = GetFriendInfo(playerNameOrID)
    else
        playerName = playerNameOrID
    end
    local wasFriend = self.isFriend(playerName)
    self.hooks.RemoveFriend(playerNameOrID)
    if wasFriend and not self.isFriend(playerName) then
        self:storeAction(playerName, REMOVE_ACTION, false)
    end
end

function SyncFriends:getPoolContainer()
    return self.db.factionrealm
end

function SyncFriends:getPool()
    local container = self:getPoolContainer()
    local pool = container.pool
    if pool == nil then
        container.pool = {}
        pool = container.pool
    end
    return pool
end

function SyncFriends:flushPool()
    self:getPoolContainer().pool = {}
    self:Print(L["Pool flushed"])
end

function SyncFriends:skipPlayerName(playerName)
    self:storeAction(playerName, SKIP_ACTION, true)
end

local COMMAND_ACTION_MAP = {
  add = ADD_ACTION,
  remove = REMOVE_ACTION,
}

function SyncFriends:registerFromCommand(playerName, action_name)
    if action_name == nil then
        action = ADD_ACTION
    else
        action = COMMAND_ACTION_MAP[action_name]
        if action == nil then
            error(string.format(L["Unknown action %s"], tostring(action_name)))
        end
    end
    self:storeAction(playerName, action, true)
end

function SyncFriends:forget(playerName)
    self:storeAction(playerName, nil, true)
end

function SyncFriends:storeAction(playerName, action, force)
    local pool = self:getPool()
    local old_value = pool[playerName]
    if (old_value ~= SKIP_ACTION or force) and old_value ~= action then
        if old_value == nil then
            self:Print(string.format(L["Marking new %s for %s"], playerName,
                       ACTION_NAME_MAP[action]))
        else
            if action == nil then
                self:Print(string.format(L["Forgeting about %s"], playerName))
            else
                self:Print(string.format(L["Marking %s for %s"], playerName,
                           ACTION_NAME_MAP[action]))
            end
        end
        pool[playerName] = action
    end
end

function SyncFriends:importExport()
    local pool = self:getPool()
    local current_playerName = UnitName("player")
    local friend_set = {}
    local storeAction
    if self:getAutoExport() then
        storeAction = self.storeAction
    else
        storeAction = function() end
    end

    -- Update pool data with friend list content
    -- Also fill a "known friend" mapping, for faster lookup when
    -- reading sync pool.
    for x = 1, GetNumFriends() do
        local playerName = GetFriendInfo(x)
        if pool[playerName] == nil then
            storeAction(self, playerName, ADD_ACTION, false)
        end
        friend_set[playerName] = true
    end
    -- Add self to pool data
    storeAction(self, current_playerName, ADD_ACTION, false)
    if self:getAutoImport() then
        -- Update friend list with pool data
        for playerName, action_type in pairs(pool) do
            -- We don't want (and cannot) make player be a friend of
            -- himself.
            if current_playerName ~= playerName then
                local action = ACTION_MAP[action_type]
                if action == nil then
                    message = string.format(L["Unknown action %s for %s, "..
                      "skipping"], tostring(action_type), playerName)
                else
                    message = action(playerName, friend_set)
                end
                if message ~= nil then
                    self:Print(message)
                end
            end
        end
    end
end

function SyncFriends:dump()
    self:Print(L["Dump start"])
    for friend_name, action in pairs(self:getPool()) do
        action_name = ACTION_NAME_MAP[action]
        if action_name == nil then
            action_name = "|cffff4444"..L["UNKNOWN ACTION"].."|r"
        end
        self:Print(friend_name..": "..action_name)
    end
end

local ui_sync_pool_action_option = {
    action = {
        name = L["Action"],
        type = "select",
        values = ACTION_NAME_MAP,
        get = "getAction",
        set = "setAction",
        style = "dropdown",
    },
    forget = {
        name = L["Forget"],
        type = "execute",
        func = "doForget",
        confirm = true,
    },
}

function UISyncPoolOptionHandlers:getAction(info)
    local friend_name = info[#info-1]
    return SyncFriends:getPool()[friend_name]
end

function UISyncPoolOptionHandlers:setAction(info, value)
    local friend_name = info[#info-1]
    SyncFriends:storeAction(friend_name, value, true)
end

function UISyncPoolOptionHandlers:doForget(info)
    local friend_name = info[#info-1]
    SyncFriends:forget(friend_name)
end

function SyncFriends:getUISyncPoolOption()
    -- Returns an option mapping generated from existing synchronisation pool,
    -- with actions possible for each entry.
    local friend_set = {}
    for friend_name, action in pairs(self:getPool()) do
        friend_set[friend_name] = {
            name = friend_name,
            type = "group",
            args = ui_sync_pool_action_option,
        }
    end
    return friend_set
end
